"""Service containing all DNS logic."""

from logging import getLogger

from certbot.errors import PluginError
from exonetapi.structures import ApiResource
from tldextract import extract

from certbot_dns_exonet.clients.exonet_client import ExonetClient

LOGGER = getLogger(__name__)


class DnsService:
    """Service containing all DNS logic."""

    client: ExonetClient

    def __init__(self, token: str) -> None:
        """DNS service constructor.

        Args:
            token: The Exonet API token.

        """
        self.client = ExonetClient(token)

    def add_txt_record(
        self, domain_name: str, record_name: str, record_content: str
    ) -> None:
        """Add a TXT record using the supplied information.

        Args:
             domain_name: The domain to use to associate the record with.
             record_name: The record name (typically beginning with '_acme-challenge.').
             record_content: The record content (typically the challenge validation).

        Raises:
             PluginError: PluginError: If an error occurs while finding DNS zone.

        """
        # Convert to registered domain.
        domain = extract(domain_name).registered_domain

        # Find the DNS zone.
        zone = self.client.find_dns_zone_by_name(domain)

        # If a zone is found, raise exception.
        if not zone:
            msg = f"Unable to find DNS zone for {domain_name}. Zone {domain} not found."
            raise PluginError(msg)

        LOGGER.debug(
            "Found DNS zone %s for domain %s", zone.attribute("name"), domain_name
        )

        LOGGER.debug("Adding TXT record to DNS.")

        # Add the TXT record to the DNS.
        record = ApiResource("dns_records")
        record.attribute("type", "TXT")
        record.attribute("name", self._compute_record_name(zone, record_name))
        record.attribute("content", self._compute_record_content(record_content))
        record.attribute("ttl", 3600)
        record.relationship("zone", zone)
        created_record = self.client.post_api_resource(record)

        LOGGER.debug("Successfully added TXT record with id: %s", created_record.id())

    def del_txt_record(
        self, domain_name: str, record_name: str, record_content: str
    ) -> None:
        """Delete a TXT record using the supplied information.

        Note that both the record's name and content are used to ensure that
        similar records created concurrently
        (e.g., due to concurrent invocations of this plugin) are not deleted.

        Failures are logged, but not raised.

        Args:
            domain_name: The domain to use to associate the record with.
            record_name: The record name (typically beginning with '_acme-challenge.').
            record_content: The record content (typically the challenge validation).

        Raises:
             PluginError: PluginError: If no DNS records are found for a domain.

        """
        # Convert to registered domain.
        domain = extract(domain_name).registered_domain

        # Find the DNS zone.
        zone = self.client.find_dns_zone_by_name(domain)

        # Get DNS records for DNS zone.
        domain_records = self.client.get_relation(zone, "records")

        # If no records are found raise exception.
        if not domain_records:
            msg = f"Unable to find DNS records for {zone}."
            raise PluginError(msg)

        # Get all matching records.
        matching_records = [
            record
            for record in domain_records
            if record.attribute("type") == "TXT"
            and record.attribute("name") == self._compute_record_name(zone, record_name)
            and record.attribute("content")
            == self._compute_record_content(record_content)
        ]

        # Delete all matching records.
        for record in matching_records:
            self.client.delete_api_resource(record)

    @staticmethod
    def _compute_record_name(domain: ApiResource, full_record_name: str) -> str:
        """Compute the DNS record name.

        Args:
            domain: The Exonet DNS record.
            full_record_name: The fill record name.

        Returns:
            str: The computed record name.

        """
        # The name of the DNS record, without the DNS zone name.
        return full_record_name.rpartition("." + domain.attribute("name"))[0]

    @staticmethod
    def _compute_record_content(record_content: str) -> str:
        """Compute the DNS record content name.

        Args:
            record_content: The record content.

        Returns:
            str: The computed record content name.

        """
        # The DNS record content within quotes.
        return '"' + record_content + '"'



"""DNS Authenticator for the Exonet API."""

from collections.abc import Callable
from logging import getLogger

from certbot.configuration import NamespaceConfig
from certbot.plugins.dns_common import CredentialsConfiguration, DNSAuthenticator

from certbot_dns_exonet.services.dns_service import DnsService

LOGGER = getLogger(__name__)


class ExonetAuthenticator(DNSAuthenticator):
    """DNS Authenticator for the Exonet API.

    This Authenticator uses the Exonet API to fulfill a dns-01 challenge.
    """

    description = "Obtain certificates using a DNS TXT record with the Exonet DNS."

    credentials: CredentialsConfiguration

    def __init__(self, config: NamespaceConfig, name: str) -> None:
        """Construct the Authenticator class.

        Args:
            config: Authenticator config.
            name: Created by name.

        """
        super().__init__(config, name)
        self._setup_credentials()

        self.dns_service = DnsService(str(self.credentials.conf("token")))

    @classmethod
    def add_parser_arguments(
        cls, add: Callable[..., None], default_propagation_seconds: int = 10
    ) -> None:
        """Add arguments to parser.

        Args:
            add: Function to call.
            default_propagation_seconds: Default propagation. Defaults to 10 seconds.

        """
        super().add_parser_arguments(add, default_propagation_seconds)
        add("credentials", help="Exonet credentials INI file.")

    def more_info(self) -> str:
        """Get more info about the plugin.

        Returns:
            Information about the plugin.

        """
        return (
            "This plugin configures a DNS TXT record to respond"
            " to a dns-01 challenge using the \
            Exonet API."
        )

    def _setup_credentials(self) -> None:
        self.credentials = self._configure_credentials(
            "credentials",
            "Exonet credentials INI file",
            {"token": "API token for Exonet API"},
        )

    def _perform(
        self, domain: str, validation_name: str, validation: str
    ) -> None:  # pragma: no cover
        """Add TXT DNS record using the Exonet API.

        Args:
            domain: The domain name.
            validation_name: The validation name.
            validation: The validation.

        """
        self.dns_service.add_txt_record(domain, validation_name, validation)

    def _cleanup(
        self, domain: str, validation_name: str, validation: str
    ) -> None:  # pragma: no cover
        """Delete TXT DNS record using the Exonet API.

        Args:
            domain: The domain name.
            validation_name: The validation name.
            validation: The validation.

        """
        self.dns_service.del_txt_record(domain, validation_name, validation)



"""Client that encapsulates all communication with the Exonet API."""

from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING

from certbot.errors import PluginError
from exonetapi import Client
from requests.exceptions import HTTPError

if TYPE_CHECKING:
    from exonetapi.structures import ApiResource, ApiResourceSet


LOGGER = getLogger(__name__)


class ExonetClient:
    """Encapsulates all communication with the Exonet API."""

    client: Client

    def __init__(self, token: str) -> None:
        """Exonet client constructor.

        Args:
            token: Exonet token.

        """
        self.client = Client()
        self.client.authenticator.set_token(token)

    def post_api_resource(self, resource: ApiResource) -> ApiResource:
        """Post the Exonet ApiResource.

        Args:
            resource: The Exonet ApiResource.

        Raises:
            PluginError: When resource can not de added.

        Returns:
            ApiResource: The created Exonet ApiResource.

        """
        try:
            return resource.post()
        except HTTPError as exception:
            description = f": {exception.response.text}" if exception.response else ""
            error_message = f"Error adding {type(resource).__name__} using the Exonet API{description}"  # noqa: E501
            LOGGER.debug(error_message)
            raise PluginError(error_message) from exception

    def delete_api_resource(self, resource: ApiResource) -> None:
        """Delete Exonet ApiResource.

        Args:
            resource: The Exonet ApiResource.

        """
        try:
            LOGGER.debug("Deleting DNS record with id: %s", resource.id())
            resource.delete()
        except HTTPError as exception:
            description = f": {exception.response.text}" if exception.response else ""
            LOGGER.warning(
                "Error deleting %s %s using the Exonet API%s",
                type(resource).__name__,
                resource.id(),
                description,
            )

    def get_relation(
        self, resource: ApiResource, relation_name: str
    ) -> ApiResourceSet | None:
        """Get relation for ApiResource based on relation name.

        Args:
            resource: The Exonet ApiResource.
            relation_name: The name of the relation.

        Returns:
            Optional[ApiResourceSet]: ApiResourceSet if found.

        """
        try:
            return resource.related(relation_name).get()
        except HTTPError as exception:
            description = f": {exception.response.text}" if exception.response else ""
            LOGGER.debug(
                "Error getting %s from the Exonet API%s",
                type(resource).__name__,
                description,
            )
            return None

    def find_dns_zone_by_name(self, domain: str) -> ApiResource | None:
        """Find the domain resource for a given domain.

        Args:
            domain: The registered domain name.

        Raises:
            PluginError: If no matching domain is found.

        Returns:
            The domain, if found.

        """
        try:
            # Get zone based on attribute name.
            zones = (
                self.client.resource("dns_zones")
                .filter("name", domain)
                .get()
                .resources()
            )
        except HTTPError as exception:
            status_code = exception.response.status_code if exception.response else None
            hint = "(Did you provide a valid API token?)" if status_code == 401 else ""
            error_message = (
                f"Error finding DNS zone using the Exonet API: {exception}{hint}"
            )
            LOGGER.debug(error_message)
            raise PluginError(error_message) from exception

        return zones[0] if zones else None
